# coding: utf-8

"""
This module defines a holder of Monosublocus instances. It is the last step
before the definition of real loci.
"""

from sys import version_info
from typing import Iterable
from ..transcripts.transcript import Transcript
from .abstractlocus import Abstractlocus
from .locus import Locus
from .monosublocus import Monosublocus
from .sublocus import Sublocus
from ..parsers.GFF import GffLine
if version_info.minor < 5:
    from sortedcontainers import SortedDict
else:
    from collections import OrderedDict as SortedDict


# Resolution order is important here!
# pylint: disable=too-many-instance-attributes
class MonosublocusHolder(Sublocus, Abstractlocus):
    """This is a container that groups together the transcripts
    surviving the selection for the Monosublocus.
    The class inherits from both sublocus and Abstractlocus
    (the main abstract class) in order to be able to reuse
    some of the code present in the former.
    Internally, the most important method is define_loci -
    which will select the best transcript(s) and remove all the overlapping ones.
    The intersection function for this object is quite laxer than in previous stages,
    and so are the requirements for the inclusion.
    """

    __name__ = "monosubloci_holder"

    # pylint: disable=super-init-not-called
    def __init__(self,
                 transcript_instance=None,
                 configuration=None, logger=None,
                 verified_introns=None, **kwargs):
        """
        :param transcript_instance:
        :type transcript_instance: Monosublocus
        :param configuration:
        :param logger:
        :param verified_introns:
        :param kwargs:
        """

        # I know what I am doing by NOT calling the Sublocus super but rather Abstractlocus
        Abstractlocus.__init__(self,
                               transcript_instance=None,
                               verified_introns=verified_introns,
                               configuration=configuration,
                               logger=logger,
                               **kwargs)
        self._not_passing = set()
        self.splitted = False
        self.metrics_calculated = False
        self.excluded = None
        self.feature = "MonosublocusHolder"
        # Add the transcript to the Locus
        self.locus_verified_introns = set()
        self.add_monosublocus(transcript_instance, check_in_locus=False)
        self.loci = SortedDict()
        self.attributes = {}

    # Overriding is correct here
    # pylint: disable=arguments-differ
    def add_transcript_to_locus(self, transcript, check_in_locus=True, **kwargs):
        """Override of the sublocus method, and reversal to the original
        method in the Abstractlocus class.
        The check_in_locus boolean flag is used to decide
        whether to check if the transcript is in the Locus or not.
        This should be set to False for the first transcript, and True afterwards.

        :param transcript: a Transcript instance
        :type transcript: Transcript

        :param check_in_locus: optional flag to pass to the basic method.
        If set to False it disables checks on whether the transcript
        is really in the locus
        :type check_in_locus: bool
        """

        if check_in_locus is True and self.in_locus(
                self,
                transcript,
                flank=self.configuration.pick.clustering.flank,
                logger=self.logger,
                cds_only=self.configuration.pick.clustering.cds_only,
                min_cdna_overlap=self.configuration.pick.clustering.min_cdna_overlap,
                min_cds_overlap=self.configuration.pick.clustering.min_cds_overlap,
                simple_overlap_for_monoexonic=self.configuration.pick.clustering.simple_overlap_for_monoexonic
        ) is False:
            self.logger.debug("%s is not a valid intersection for %s", transcript.id, self.id)
            return False

        Abstractlocus.add_transcript_to_locus(self, transcript, check_in_locus=False)
        self.locus_verified_introns = set.union(self.locus_verified_introns,
                                                transcript.verified_introns)

    # pylint: enable=arguments-differ
    @classmethod
    def is_intersecting(cls,
                        transcript,
                        other,
                        cds_only=False,
                        logger=None,
                        min_cdna_overlap=0.2,
                        min_cds_overlap=0.2,
                        simple_overlap_for_monoexonic=True) -> bool:
        """Alias for the Sublocus.is_intersecting method"""

        return Sublocus.is_intersecting(transcript, other, cds_only=cds_only,
                                        logger=logger, min_cdna_overlap=min_cdna_overlap,
                                        min_cds_overlap=min_cds_overlap,
                                        simple_overlap_for_monoexonic=simple_overlap_for_monoexonic)

    def add_monosublocus(self, monosublocus_instance: Monosublocus,
                         check_in_locus=True):
        """Wrapper to extract the transcript from the monosubloci and pass it to the constructor.

        :param monosublocus_instance: the instance to add
        :type monosublocus_instance: Monosublocus

        :param check_in_locus: boolean flag - should we perform checks to verify the monosublocus is compatible?
        """
        if monosublocus_instance is None:
            return

        assert len(monosublocus_instance.transcripts) == 1
        for tid in monosublocus_instance.transcripts:
            self.add_transcript_to_locus(monosublocus_instance.transcripts[tid],
                                         check_in_locus=check_in_locus)

    def __str__(self, print_cds=False, source_in_name=True):
        """
        Method to print out the sublocus lines in GFF format.
        :param print_cds: flag. Should we print the CDS?
        :param source_in_name: boolean flag. Should we add the original source of the transcript to the ID?
        """

        lines = []

        self_line = GffLine('')
        for attr in ["chrom", 'feature', 'source', 'start', 'end', 'strand']:
            setattr(self_line, attr, getattr(self, attr))
        self.filter_and_calculate_scores()

        self_line.phase, self_line.score = None, self.score
        if source_in_name is True:
            self_line.id = "{0}_{1}".format(self.source, self.id)
        else:
            self_line.id = self.id
        self_line.name = self.name
        self_line.parent = self.parent
        self_line.attributes.update(self.attributes)
        self_line.attributes["multiexonic"] = (not self.monoexonic)
        lines.append(str(self_line))

        for tid in self.transcripts:
            transcript_instance = self.transcripts[tid]
            transcript_instance.source = self.source
            transcript_instance.parent = self_line.id
            self.logger.debug(self.attributes)
            for attribute in self.attributes:
                if attribute not in transcript_instance.attributes:
                    if attribute == "is_fragment" and self.attributes[attribute] is False:
                        continue
                    transcript_instance.attributes[attribute] = self.attributes[attribute]

            lines.append(transcript_instance.format(
                "gff",
                all_orfs=self.configuration.pick.output_format.report_all_orfs,
                with_cds=print_cds).rstrip())

        return "\n".join(lines)

    def define_monosubloci(self, **kwargs):
        """Overriden and set to NotImplemented to avoid cross-calling it when inappropriate."""
        raise NotImplementedError("Monosubloci are the input of this object, not the output.")

    def define_loci(self, purge=False, excluded=None, check_requirements=True):
        """This is the main function of the class. It is analogous
        to the define_subloci class defined for sublocus objects,
        but it returns "Locus" objects (not "Monosublocus").

        Optional parameters:

        :param purge: flag. If set to True, all loci whose transcripts
        have scores of 0 will be thrown out
        into an Excluded holder.
        :type purge: bool

        :param excluded
        :type excluded: Excluded

        :param check_requirements: boolean flag. If set to false, transcripts will not be checked for passing requirements.
        :type check_requirements: bool

        """
        if self.splitted is True:
            return

        self.excluded = excluded

        self.filter_and_calculate_scores(check_requirements=check_requirements)

        graph = self.define_graph(
            self.transcripts,
            inters=self.is_intersecting,
            logger=self.logger,
            cds_only=self.configuration.pick.clustering.cds_only,
            min_cdna_overlap=self.configuration.pick.clustering.min_cdna_overlap,
            min_cds_overlap=self.configuration.pick.clustering.min_cds_overlap,
            simple_overlap_for_monoexonic=self.configuration.pick.clustering.simple_overlap_for_monoexonic
        )

        loci = []
        while len(graph) > 0:
            communities = self.find_communities(graph)
            to_remove = set()
            for locus_comm in communities:
                locus_comm = {x: self.transcripts[x] for x in locus_comm}
                selected_tid = self.choose_best(locus_comm)
                selected_transcript = self.transcripts[selected_tid]
                to_remove.add(selected_tid)
                to_remove.update(set(graph.neighbors(selected_tid)))
                if purge is False or selected_transcript.score > 0:
                    new_locus = Locus(selected_transcript, logger=self.logger, configuration=self.configuration,
                                      use_transcript_scores=self._use_transcript_scores)
                    loci.append(new_locus)

            self.logger.debug("Removing {0} transcripts from {1}".format(len(to_remove), self.id))
            graph.remove_nodes_from(to_remove)  # Remove nodes from graph, iterate

        for locus in sorted(loci):
            self.loci[locus.id] = locus
        self.splitted = True
        return

    def add_transcripts_to_locus(
            self, transcripts: Iterable[Transcript],
            check_in_locus=True, overwrite=False,
            **kwargs):

        for transcript in transcripts:
            self.add_transcript_to_locus(transcript, check_in_locus=check_in_locus,
                                         **kwargs)

    @classmethod
    def in_locus(cls, monosublocus: Abstractlocus,
                 transcript: Transcript,
                 flank=0,
                 logger=None,
                 cds_only=False,
                 min_cdna_overlap=0.2,
                 min_cds_overlap=0.2,
                 simple_overlap_for_monoexonic=False) -> bool:

        """This method checks whether a transcript / monosbulocus
        falls inside the Locus coordinates.

        :rtype: bool

        :param monosublocus: Monosublocus instance
        :type monosublocus: Monosublocus

        :param transcript: the transcript to be compared
        :type transcript: Transcript

        :param flank: optional flank argument
        :type flank: int

        :param logger: logger to use (one will be created if missing)
        :type logger: Logger

        :param cds_only: whether to consider only the CDS part (if the transcript is coding) for
                         the evaluation.
        :type cds_only: bool

        :param min_cds_overlap: minimum CDS overlap for the transcript to fit in the locus.
        :type min_cds_overlap: float

        :param min_cdna_overlap: minimum cDNA overlap for the transcript to fit in the locus.
        :type min_cdna_overlap: float

        :param simple_overlap_for_monoexonic: if true, consider a simple overlap for monoexonic transcripts.
        :type simple_overlap_for_monoexonic: bool
        """

        if hasattr(transcript, "transcripts"):
            assert len(transcript.transcripts) == 1
            transcript = transcript.transcripts[list(transcript.transcripts.keys())[0]]
            assert hasattr(transcript, "finalize")
        if Abstractlocus.in_locus(monosublocus, transcript, flank=flank) is True:
            is_in_locus = False
            for tran in monosublocus.transcripts:
                tran = monosublocus.transcripts[tran]
                is_in_locus = cls.is_intersecting(tran,
                                                  transcript,
                                                  logger=logger,
                                                  cds_only=cds_only,
                                                  min_cds_overlap=min_cds_overlap,
                                                  min_cdna_overlap=min_cdna_overlap,
                                                  simple_overlap_for_monoexonic=simple_overlap_for_monoexonic
                                                  )
                if is_in_locus is True:
                    break
            return is_in_locus
        else:
            return False

    @property
    def id(self):
        """
        Wrapper for the id method of abstractlocus. Necessary to redefine the name.
        """

        return Abstractlocus.id.fget(self)  # @UndefinedVariable

    def as_dict(self):
        return Abstractlocus.as_dict(self)

    def load_dict(self, state, load_transcripts=True, load_configuration=True):
        Abstractlocus.load_dict(self, state, load_configuration=load_configuration,
                                load_transcripts=load_transcripts)



import unittest
import msgpack
import zlib
from ..preparation import checking
from .. import utilities
from .. import transcripts
import multiprocessing as mp
import pkg_resources
import tempfile
import logging
import logging.handlers
from pytest import mark
import pickle
import os
import time
import pyfaidx
import pysam
import re
from ..tests.test_utils import ProcRunner
from queue import Queue
from sys import version_info


class MiscTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):

        cls.fasta = pkg_resources.resource_filename("Mikado.tests", "chr5.fas.gz")
        cls.fai = pysam.FastaFile(cls.fasta)

    @staticmethod
    def create_logger(name, simple=True):
        if simple is True:
            logging_queue = Queue()
            logging_queue.put_nowait = logging_queue.put
        else:
            logging_queue = mp.JoinableQueue(-1)
        log_queue_handler = logging.handlers.QueueHandler(logging_queue)
        log_queue_handler.setLevel(logging.DEBUG)

        logger = utilities.log_utils.create_null_logger(name, level="WARNING")
        logger.propagate = False
        listener = logging.handlers.QueueListener(logging_queue, logger)
        listener.propagate = False
        listener.start()
        return logger, listener, logging_queue

    def setUp(self):
        # Create the queues for logging and submission
        self.submission_queue = mp.SimpleQueue()
        self.fasta_out = "temporary.fasta"
        self.gtf_out = "temporary.gtf"

    @unittest.skipUnless((version_info.minor > 5 and version_info.major == 3),
                         "SimpleQueues do not function properly for this in python 3.4 and earlier")
    def test_normal(self):
        logger, listener, logging_queue = self.create_logger("test_normal")

        with self.assertLogs(logger=logger, level="DEBUG") as cmo, \
                tempfile.NamedTemporaryFile(mode="wb", delete=True) as batch_file:
            # FASTA out and GTF out are just the file names, without the temporary directory
            # Moreover they will be complemented by the identifier!

            proc = ProcRunner(checking.CheckingProcess,
                              batch_file.name,
                              logging_queue,
                              shelve_stacks=[],
                              fasta=self.fasta,
                              identifier=0,
                              fasta_out=self.fasta_out,
                              gtf_out=self.gtf_out,
                              seed=None,
                              tmpdir=tempfile.gettempdir(),
                              log_level="DEBUG")
            msgpack.dump([], batch_file)
            batch_file.flush()
            proc.start()
            time.sleep(0.1)  # Necessary otherwise the check might be too fast for the FileSystem
            self.assertEqual(proc.func.fasta_out, os.path.join(tempfile.gettempdir(), self.fasta_out + "-0"))
            self.assertTrue(os.path.exists(proc.func.fasta_out), proc.func.fasta_out)
            self.assertEqual(proc.func.gtf_out, os.path.join(tempfile.gettempdir(), self.gtf_out + "-0"))
            self.assertTrue(os.path.exists(proc.func.gtf_out), proc.func.gtf_out)
            self.submission_queue.put(("EXIT", None, None, None))
            time.sleep(0.1)
            proc.stop()
            os.remove(proc.func.fasta_out)
            os.remove(proc.func.gtf_out)
            batch_file.close()
            assert not proc.is_alive()

        self.maxDiff = 10000
        # self.assertEqual(cmo.output, [
        #     "DEBUG:Checker-0:Starting Checker-0",
        #     "DEBUG:Checker-0:Created output FASTA {} and GTF {}".format(proc.func.fasta_out, proc.func.gtf_out),
        #     "DEBUG:Checker-0:(('GT', 'AG'), ('GC', 'AG'), ('AT', 'AC'))",
        #     "DEBUG:Checker-0:Finished for Checker-0"])

        self.assertIsInstance(proc.func, mp.Process)

        with self.assertRaises(TypeError):
            _ = pickle.dumps(proc.func)

    def test_wrong_initialisation(self):

        if version_info.minor < 6:
            self.submission_queue = mp.Queue(-1)
            logger, listener, logging_queue = self.create_logger("test_wrong_initialisation", simple=False)
        else:
            logger, listener, logging_queue = self.create_logger("test_wrong_initialisation", simple=True)

        with tempfile.NamedTemporaryFile() as batch_file, tempfile.NamedTemporaryFile() as shelves:
            pass

        kwds = {"batch_file": batch_file.name,
                "logging_queue": logging_queue,
                "identifier": 0,
                "shelve_stacks": [shelves.name],
                "fasta_out": self.fasta_out,
                "gtf_out": self.gtf_out,
                "tmpdir": tempfile.gettempdir(),
                "fasta": self.fasta,
                "seed": None,
                "log_level": "WARNING"
                }

        for key in ["batch_file", "logging_queue", "identifier", "lenient"]:
            with self.subTest(key=key):
                _kwds = kwds.copy()
                _kwds[key] = None
                if "queue" in key:
                    error = TypeError
                else:
                    error = ValueError
                with self.assertRaises(error):
                    checking.CheckingProcess(**_kwds)

        with self.subTest(key="fasta"):
            _kwds = kwds.copy()
            _kwds["fasta"] = None

            with self.assertRaises((AttributeError, TypeError)):
                checking.CheckingProcess(**_kwds)

        for tentative in [None, [], [("A", "G")], [("AG", bytes("GT", encoding="ascii"))]]:
            with self.subTest(tentative=tentative):
                _kwds = kwds.copy()
                _kwds["canonical_splices"] = tentative
                if tentative is None:
                    with self.assertRaises(TypeError):
                        checking.CheckingProcess(**_kwds)
                else:
                    with self.assertRaises(ValueError):
                        checking.CheckingProcess(**_kwds)

        _kwds = kwds.copy()
        _kwds["canonical_splices"] = [("AG", "GT")]
        # just test it does not raise
        _ = checking.CheckingProcess(**_kwds)

    @mark.slow
    def test_example_model(self):

        fasta = pyfaidx.Fasta(self.fasta)
        lines = dict()
        lines["chrom"] = "Chr5"
        lines["strand"] = "+"
        lines["start"] = 208937
        lines["end"] = 210445
        lines["attributes"] = dict()
        lines["tid"], lines["parent"] = "AT5G01530.0", "AT5G01530"
        lines["features"] = dict()
        lines["features"]["exon"] = [(208937, 209593), (209881, 210445)]
        seq = str(fasta[lines["chrom"]][lines["start"] - 1:lines["end"]])

        logger, listener, logging_queue = self.create_logger("test_example_model")

        res = checking.create_transcript(lines, seq, lines["start"], lines["end"],
                                                            logger=logger)
        listener.stop()
        self.assertIsInstance(res, transcripts.TranscriptChecker)

        for kwd in lines.keys():
            if kwd in ["end", "start"]:
                continue
            with self.subTest(kwd=kwd, msg="Testing key {}".format(kwd)):
                _lines = lines.copy()
                del _lines[kwd]
                with self.assertLogs("null", level="DEBUG"):
                    res = checking.create_transcript(_lines, seq, lines["start"], lines["end"])
                self.assertIs(res, None)

        _lines = lines.copy()
        _lines["strand"] = "-"
        with self.subTest(msg="Testing an invalid strand"):
            with self.assertLogs("null", level="INFO") as cm:
                res = checking.create_transcript(_lines, seq, lines["start"], lines["end"],
                                                                    strand_specific=True)
            self.assertIsInstance(res, transcripts.TranscriptChecker)
            self.assertIn("WARNING:null:Transcript AT5G01530.0 has been assigned to the wrong strand, reversing it.",
                          cm.output)

    @unittest.skipUnless((version_info.minor > 5 and version_info.major == 3),
                         "SimpleQueues do not function properly for this in python 3.4 and earlier")
    def test_example_model_through_process(self):

        logger, listener, logging_queue = self.create_logger("test_example_model_through_process")
        logger.setLevel("DEBUG")
        with self.assertLogs(logger=logger, level="DEBUG") as cmo, \
                tempfile.NamedTemporaryFile(delete=True, suffix=".db", mode="wb") as dumpdb, \
                tempfile.NamedTemporaryFile(delete=True, mode="wb") as batch:
            # FASTA out and GTF out are just the file names, without the temporary directory
            # Moreover they will be complemented by the identifier!
            lines = dict()
            lines["chrom"] = "Chr5"
            lines["strand"] = "+"
            lines["start"] = 208937
            lines["end"] = 210445
            lines["attributes"] = dict()
            lines["tid"], lines["parent"] = "AT5G01530.0", "AT5G01530"
            lines["features"] = dict()
            lines["features"]["exon"] = [(208937, 209593), (209881, 210445)]
            lines["strand_specific"] = True
            lines["is_reference"] = False
            write_start = dumpdb.tell()
            dumpdb.write(zlib.compress(msgpack.dumps(lines)))
            write_length = dumpdb.tell() - write_start
            dumpdb.flush()

            keys = [(0, ((lines["tid"], dumpdb.name, write_start, write_length),
                         lines["chrom"], (lines["start"], lines["end"])))]
            msgpack.dump(keys, batch)
            batch.flush()
            proc = ProcRunner(checking.CheckingProcess,
                              batch.name,
                              logging_queue,
                              fasta=self.fasta,
                              shelve_stacks=[dumpdb.name],
                              identifier=logger.name,
                              fasta_out=self.fasta_out,
                              gtf_out=self.gtf_out,
                              seed=None,
                              tmpdir=tempfile.gettempdir(),
                              strip_faulty_cds=True,
                              log_level="DEBUG")

            self.submission_queue.put((lines, lines["start"], lines["end"], 0))
            self.submission_queue.put(("EXIT", None, None, None))
            proc.start()
            proc.join()
            time.sleep(0.5)

        self.assertTrue(os.stat(proc.func.fasta_out).st_size > 0, (proc.func.fasta_out,
                        cmo.output))
        fasta_lines = []
        with open(proc.func.fasta_out) as f_out:
            for line in f_out:
                line = line.rstrip()
                line = re.sub("0/", "", line)
                fasta_lines.append(line)

        self.assertGreater(len(fasta_lines), 1)
        seq = self.fai.fetch(lines["chrom"], lines["start"] - 1, lines["end"])
        res = checking.create_transcript(lines, seq, lines["start"], lines["end"])
        self.assertTrue(len(res.cdna), (209593 - 208937 + 1) + (210445 - 209881 + 1))

        with tempfile.NamedTemporaryFile(suffix="fa", delete=True, mode="wt") as faix:

            assert len(fasta_lines) > 1
            assert fasta_lines[0][0] == ">"
            for line in fasta_lines:
                print(line, file=faix)
            faix.flush()
            try:
                fa = pyfaidx.Fasta(faix.name)
            except TypeError:
                raise TypeError([_ for _ in open(faix.name)])
            self.assertEqual(list(fa.keys()), ["AT5G01530.0"])
            self.assertEqual(str(fa["AT5G01530.0"]), str(res.cdna))
            self.assertEqual(len(str(fa["AT5G01530.0"])), res.cdna_length)

            os.remove(faix.name + ".fai")

        listener.stop()

    def test_lowercase_ref_splice_site(self):
        reference = pkg_resources.resource_filename("Mikado.tests", "NC_037283.1.fa.gz")
        fasta = pyfaidx.Fasta(reference)
        fai = pysam.FastaFile(reference)

        lines = dict()
        lines["chrom"] = "NC_037283.1"
        lines["strand"] = '-'
        lines["start"] = 200431
        lines["end"] = 204262
        lines["attributes"] = dict()
        lines["tid"], lines["parent"] = "STRG.4616.1", "STRG.4616"
        lines["features"] = dict()
        lines["features"]["exon"] = [(200431, 200919), (201096, 201282), (201446, 201512),
                                     (201776, 203421), (203570, 204262)]

        seq = str(fasta[lines["chrom"]][lines["start"] - 1:lines["end"]])

        logger, listener, logging_queue = self.create_logger("test_example_model")

        res = checking.create_transcript(lines, seq, lines["start"], lines["end"],
                                         logger=logger)
        listener.stop()
        self.assertIsInstance(res, transcripts.TranscriptChecker)
        self.assertEqual(res.attributes["canonical_number"], 4)



import sys
import argparse
import tabulate
from ...scales import class_codes
import textwrap
from itertools import zip_longest


def launch(args):

    rows = []

    if len(args.code) > 0:
        codes = [class_codes.codes[_] for _ in args.code]
    elif len(args.category) > 0:
        codes = [class_codes.codes[_] for _ in class_codes.codes
                 if class_codes.codes[_].category in args.category]
    else:
        codes = [class_codes.codes[_] for _ in class_codes.codes]

    for code in codes:
        definition = textwrap.wrap(code.definition, 30)
        code_rows = zip_longest([code.code],
                                definition,
                                [code.ref_multi],
                                [code.pred_multi],
                                [code.nucl],
                                [code.junc],
                                [code.reverse],
                                code.category.split())
        rows.extend(code_rows)

    __table_format = tabulate._table_formats[args.format]

    if args.format not in ("grid", "fancy_grid"):

        print(tabulate.tabulate(rows,
                                headers=["Class code",
                                         "Definition",
                                         "Reference multiexonic?",
                                         "Prediction multiexonic?",
                                         "Nucleotide: RC, PC, F1",
                                         "Junction: RC, PC, F1",
                                         "Reverse",
                                         "Category"],
                                tablefmt=args.format))

    else:
        out_of_header = False
        separator = None

        for row in tabulate.tabulate(rows,
                                     headers=["Class code",
                                              "Definition",
                                              "Reference multiexonic?",
                                              "Prediction multiexonic?",
                                              "Nucleotide: RC, PC, F1",
                                              "Junction: RC, PC, F1",
                                              "Reverse",
                                              "Category"],
                                     tablefmt=args.format).split("\n"):
            if row[:2] == __table_format.lineabove.begin + __table_format.lineabove.hline:
                separator = row
                if not out_of_header:
                    print(row)
                continue
            if row[:2] == __table_format.linebelowheader.begin + __table_format.linebelowheader.hline:
                out_of_header = True
                print(row, file=args.out)
                continue
            elif out_of_header is False:
                print(row, file=args.out)
            elif row[:2] == __table_format.linebetweenrows[0] + __table_format.linebetweenrows[1]:
                continue
            elif row[0] == __table_format.datarow.begin:
                if row.strip().split(__table_format.datarow.sep)[1].strip() != "":
                    print(separator, file=args.out)
                print(row, file=args.out)
        print(separator, file=args.out)
        print(file=args.out)
    return


def code_parser():

    """
    Command line parser for the utility.
    """

    parser = argparse.ArgumentParser("Script to generate the available class codes.")
    parser.add_argument("-f", "--format", choices=tabulate.tabulate_formats, default="rst")
    parser.add_argument("-c", "--category", nargs="+", default=[],
                        choices=list(set(_.category for _ in class_codes.codes.values())))
    parser.add_argument("-o", "--out", type=argparse.FileType("w"), default=sys.stdout)
    parser.add_argument("code", nargs="*", help="Codes to query.",
                        default=[],
                        choices=[[]] + list(class_codes.codes.keys()))
    parser.set_defaults(func=launch)
    return parser







#!/usr/bin/env python3

import Mikado
import argparse
# import sys
import collections
from Mikado.utilities.intervaltree import IntervalTree
# import os
import subprocess
import multiprocessing
import tempfile
from scipy.stats.mstats import mquantiles
from numpy import arange

from Mikado.scales.reference_preparation.indexing import load_index
from Mikado.utilities import log_utils
from Mikado.utilities.log_utils import LoggingConfiguration


class Geomancer(multiprocessing.Process):

    def __init__(self, gid_pipe: multiprocessing.Pipe,
                 positions: dict,
                 out_file: str,
                 index: dict):

        super(Geomancer, self).__init__()
        self.gid_pipe = gid_pipe
        self.index = index
        self.out = open(out_file, "wt")
        self.positions = positions

    def run(self):

        while True:

            # This part basically replicates the Mikado.scales.assigner.Assigner.find_neighbours function

            (gid, chrom, start, end, strand) = self.gid_pipe.get()
            if gid is None:
                self.gid_pipe.put((None, None, None, None, None))
                self.out.flush()
                self.out.close()
                return

            found = self.index[chrom].search(start, end, max_distance=10**8, num_intervals=1, strict=False)
            if not found:
                line = [gid, "{}:{}..{}".format(chrom, start, end)] + ["NA"] * 6
                self.out.write("\t".join(str(_) for _ in line) + "\n")
                continue

            before, after = [], []

            for key in found:
                key = (key.start, key.end)
                if key[1] < start:
                    before.append(key)
                elif key[0] > end:
                    after.append(key)
                else:
                    found_gids = [_ for _ in self.positions[chrom][key] if _ != gid]
                    if found_gids:
                        before.append(key)
                        after.append(key)
                    else:
                        continue

            before = sorted(before)
            after = sorted(after)

            if len(before) == 0:
                gid_before = "NA"
                position_before = "NA"
                distance_before = "NA"
            else:
                gid_before = ",".join(_ for _ in self.positions[chrom][before[-1]])
                position_before = "{}:{}..{}".format(chrom, *before[-1])
                distance_before = max(0, start - before[-1][1] + 1)
            if len(after) == 0:
                gid_after = "NA"
                distance_after = "NA"
                position_after = "NA"
            else:
                gid_after = ",".join(_ for _ in self.positions[chrom][after[0]])
                distance_after = max(0, after[0][0] - end + 1)
                position_after = "{}:{}..{}".format(chrom, *after[0])

            line = [gid, "{}:{}..{}".format(chrom, start, end),
                    gid_before, distance_before, position_before,
                    gid_after, distance_after, position_after]

            self.out.write("\t".join(str(_) for _ in line) + "\n")
            continue

    def join(self, timeout=None):
        # self.__close_handles()
        # self.terminate()
        super(Geomancer, self).join(timeout=timeout)


def main():

    parser = argparse.ArgumentParser(__doc__)
    parser.add_argument("-o", "--out", type=str, default="distances.txt")
    parser.add_argument("-l", "--log", default=None)
    parser.add_argument("-lv", "--log-level", default="INFO", choices=["DEBUG", "INFO", "WARN", "ERROR", "CRITICAL"],
                        dest="log_level")
    parser.add_argument("-p", "--procs", type=int, default=1)
    parser.add_argument("gff3")

    args = parser.parse_args()

    conf = LoggingConfiguration(log=args.log, log_level=args.log_level)
    logger = log_utils.create_logger_from_conf(conf, "geomancer", mode="w")
    logger.info("Starting to load the GFF3 index")
    with open(args.gff3) as gff3:
        namespace = argparse.Namespace
        namespace.reference = gff3
        namespace.exclude_utr = False
        namespace.protein_coding = False
        # Use Mikado compare functions to load the index from the GFF3
        # "genes" is a dictionary of Gene objects, having as keys the gene names
        # "positions" is a dictionary of the form: [chrom][(start, end)] = [GID1, GID2, ...]
        genes, positions = load_index(args, logger)
        # Create a dictionary of interval trees, one per chromosome
        indexer = collections.defaultdict(list).fromkeys(positions)
        for chrom in indexer:
            indexer[chrom] = IntervalTree.from_tuples(positions[chrom].keys())
    logger.info("Loaded the index")

    # Create different processes
    gid_pipe = multiprocessing.JoinableQueue()
    # res_pipe =

    procs = []
    logger.info("Creating children processes")
    for num in range(args.procs):
        temp = tempfile.NamedTemporaryFile(mode="w+t", delete=False)
        temp.close()
        _ = Geomancer(gid_pipe, positions, temp.name, indexer)
        _.start()
        procs.append(_)
    logger.info("Created children processes")

    logger.info("Starting to analyse {} genes".format(len(genes)))
    # Lookup in a small set is MUCH faster than lookup in a small np.array
    steps = set(mquantiles(range(1, len(genes.keys())), prob=arange(0.1, 1.1, 0.1)).astype(int))

    for counter, gid in enumerate(genes.keys(), 1):
        if counter in steps:
            logger.info("Analysed {} genes ({}%)".format(counter, int(counter * 100/len(genes))))
        chrom, start, end, strand = genes[gid].chrom, genes[gid].start, genes[gid].end, genes[gid].strand
        gid_pipe.put((gid, chrom, start, end, strand))

    gid_pipe.put((None, None, None, None, None))
    # [_.join() for _ in procs]

    with open(args.out, "wt") as out:
        out.close()
        for proc in procs:
            subprocess.call("cat {} >> {}".format(proc.out.name, args.out), shell=True)
            # subprocess.call("rm {}".format(proc.out.name), shell=True)

main()



import abc
from ..utilities.file_type import filetype
import io
import gzip
import bz2
from functools import partial


class HeaderError(Exception):
    """
    Mock exception which is raised when a header/comment line (e.g. starting with "#") is found.
    """
    pass


class Parser(metaclass=abc.ABCMeta):
    """Generic parser iterator. Base parser class."""

    def __init__(self, handle):
        self.__closed = False
        self._handle = self.__get_handle(handle)
        self.closed = False

    def __iter__(self):
        return self

    def __get_handle(self, handle, position=None):
        if not isinstance(handle, io.IOBase):
            if handle.endswith(".gz") or filetype(handle) == b"application/gzip":
                opener = gzip.open
            elif handle.endswith(".bz2") or filetype(handle) == b"application/x-bzip2":
                opener = bz2.open
            else:
                opener = partial(open, **{"buffering": 1})
            try:
                handle = opener(handle, "rt")
            except FileNotFoundError:
                raise FileNotFoundError("File not found: {0}".format(handle))
        if position is not None:
            handle.seek(position)
        return handle

    def __next__(self):
        return next(self._handle)

    def __enter__(self):
        if self.closed is True:
            raise ValueError('I/O operation on closed file.')
        return self

    def __exit__(self, *args):
        _ = args
        self._handle.close()
        self.closed = True

    def close(self):
        """
        Alias for __exit__
        """
        self.__exit__()

    @property
    def name(self):
        """
        Return the filename.
        """
        if hasattr(self._handle, "name"):
            return self._handle.name
        return None

    @property
    def closed(self):
        """
        Boolean flag. If True, the file has been closed already.
        """
        return self.__closed

    @closed.setter
    def closed(self, *args):
        """
        :param args: boolean flag

        This sets the closed flag of the file.

        """
        if not isinstance(args[0], bool):
            raise TypeError("Invalid value: {0}".format(args[0]))

        self.__closed = args[0]

    def __getstate__(self):
        try:
            position = self._handle.tell()
        except:
            position = None
        state = dict()
        state.update(self.__dict__)
        state["position"] = position
        if hasattr(self._handle, "filename"):
            _handle = self._handle.filename
            if isinstance(_handle, bytes):
                _handle = _handle.decode()
            state["_handle"] = _handle
        elif hasattr(self._handle, "name"):
            _handle = self._handle.name
            if isinstance(_handle, bytes):
                _handle = _handle.decode()
            state["_handle"] = _handle
        else:
            raise TypeError("Unknown handle: {}".format(self._handle))
        state.pop("logger", None)
        return state

    def __setstate__(self, state):
        position = state.get("position")
        del state["position"]
        self.__dict__.update(state)
        self._handle = self.__get_handle(state["_handle"], position=position)


def _attribute_definition(val):
    try:
        val = float(val)
        if val.is_integer():
            return int(val)
        return val
    except (ValueError, TypeError):
        if val.lower() in ("true", "false"):
            val = val.capitalize()
            if val == "True":
                return True
            else:
                return False
        return val




# coding: utf-8

"""
This class is the main workhorse of the compare.py utility.
"""

import argparse
import collections
import csv
import gzip
import logging
import operator
import queue
import sys
from collections import namedtuple
from functools import partial
from logging import handlers as log_handlers
from ...transcripts.transcript import Transcript
from ...utilities.namespace import Namespace
from ..accountant import Accountant
from ..contrast import compare as c_compare
from ..resultstorer import ResultStorer
from ...exceptions import InvalidTranscript, InvalidCDS
from ...utilities import IntervalTree
import msgpack
import tempfile
from ..reference_preparation.gene_dict import GeneDict


# noinspection PyPropertyAccess,PyPropertyAccess

def msgpack_default(o):
    """Function to convert the types for msgpack"""
    if isinstance(o, tuple):
        return {'__type__': 'tuple', 'value': list(o)} 
    elif isinstance(o, set):
        return {'__type__': 'set', 'value': list(o)}
    elif isinstance(o, ResultStorer):
        return {'__type__': 'rstor', 'value': o.as_dict()}
    else:
        return o


def msgpack_convert(o):
    """Function to re-extract the data from msgpack"""
    if isinstance(o, dict) and o.get('__type__', None):
        if o['__type__'] == "set":
            return set(o['value'])
        elif o['__type__'] == b'list':
            return list(o[b'value'])
        elif o['__type__'] == 'tuple':
            return tuple(o['value'])
        elif o['__type__'] == 'rstor':
            return ResultStorer(state=o['value'])
        else:
            return o['value']
    else:
        return o


class Assigner:

    """
    This class has the purpose of assiging each prediction transcript to its best match
    among the reference transcripts.
    """

    def __init__(self,
                 index: str,
                 args: argparse.Namespace,
                 results=(),
                 printout_tmap=True,
                 fuzzymatch=0,
                 counter=None):

        """

        :param genes: a dictionary which contains
        the gene containers for the reference transcript objects.
        :type genes: dict

        :param positions: a defaultdict which is used for fast lookup of genomic positions
        :type positions: collections.defaultdict

        :param args: the parameters passed through the command line
        :type args: None | argparse.Namespace

        :param stat_calculator: an instance of Accountant, to which the results will be sent to.
        :type stat_calculator: Accountant

        :param printout_tmap: boolean value. If set to True, the object will print each row.
        :type printout_tmap: bool
        """

        if args is None:
            self.args = argparse.Namespace()
            self.args.out = sys.stdout
            self.args.distance = 2000
            self.args.protein_coding = False
            self.args.loq_queue = queue.Queue()
            self.args.exclude_utr = False
            self.args.verbose = False
            self.lenient = False
            self.use_prediction_alias = False
            self.__report_fusions = True
        else:
            self.args = args
            if not hasattr(args, "out"):
                self.args.out = sys.stdout
            if not hasattr(args, "distance"):
                self.args.distance = 2000
            if not hasattr(args, "protein_coding"):
                self.args.protein_coding = False
            if not hasattr(args, "loq_queue"):
                self.args.loq_queue = queue.Queue()
            if not hasattr(args, "exclude_utr"):
                self.args.exclude_utr = False
            if not hasattr(args, "verbose"):
                self.args.verbose = False
            self.lenient = getattr(args, "lenient", False)
            self.use_prediction_alias = getattr(args, "use_prediction_alias", False)
            self.__report_fusions = getattr(args, "report_fusions", True)

        # noinspection PyUnresolvedReferences
        # pylint: disable=no-member
        self.queue_handler = log_handlers.QueueHandler(self.args.log_queue)

        # pylint: enable=no-member
        if counter is None:
            self.logger = logging.getLogger("Assigner")
        else:
            self.logger = logging.getLogger("Assigner-{}".format(counter))
        self.logger.addHandler(self.queue_handler)
        # noinspection PyUnresolvedReferences
        if args.verbose:
            self.logger.setLevel(logging.DEBUG)
        else:
            self.logger.setLevel(logging.INFO)

        if hasattr(args, "fuzzymatch") and args.fuzzymatch not in (0, None):
            if fuzzymatch in (0, None):
                fuzzymatch = args.fuzzymatch

        if not isinstance(fuzzymatch, int) and not (fuzzymatch is None):
            raise TypeError(fuzzymatch)

        self.__fuzzymatch = fuzzymatch

        self.logger.propagate = True
        self.dbname = index
        self.genes = GeneDict(self.dbname, logger=self.logger,
                              exclude_utr=self.args.exclude_utr, protein_coding=self.args.protein_coding)
        self.positions = collections.defaultdict(dict)
        self.indexer = collections.defaultdict(list)
        self._load_positions()
        self.printout_tmap = printout_tmap
        self.__done = 0
        # noinspection PyUnresolvedReferences
        if self.printout_tmap:
            if self.args.gzip is False:
                self.tmap_out = open("{0}.tmap".format(args.out), 'wt')
            else:
                self.tmap_out = gzip.open("{0}.tmap.gz".format(args.out), 'wt')
            self.tmap_rower = csv.DictWriter(self.tmap_out, ResultStorer.__slots__, delimiter="\t")
            self.tmap_rower.writeheader()

        self.gene_matches = collections.defaultdict(dict)
        self.done = 0

        if self.printout_tmap is True:
            for gid in self.genes:
                for tid in self.genes[gid].transcripts:
                    self.gene_matches[gid][tid] = []

        self.stat_calculator = Accountant(self.genes, args=args,
                                          fuzzymatch=self.__fuzzymatch, counter=counter,
                                          load_ref=self.printout_tmap)
        self.self_analysis = self.stat_calculator.self_analysis
        self.__merged = False

        if results:
            self.load_from_results(results)
            self.__merged = True
            return

    def _load_positions(self):

        for row in self.genes.positions:
            chrom, start, end, gid = row
            key = (start, end)
            if key not in self.positions[chrom]:
                self.positions[chrom][key] = []
            self.positions[chrom][key].append(gid)

        for chrom in self.positions:
            self.indexer[chrom] = IntervalTree.from_tuples(self.positions[chrom].keys())

    def load_result(self, refmap, stats):

        gene_matches = msgpack.loads(refmap,
                                     raw=False, use_list=False,
                                     strict_map_key=False, object_hook=msgpack_convert)

        for gid, gene_match in gene_matches.items():
            for tid in gene_match:
                for match in gene_match[tid]:
                    self.gene_matches[gid][tid].append(ResultStorer(state=match))

        temp_stats = Namespace()
        for attribute, stat in stats:
            stat = msgpack.loads(stat,
                                 raw=False,
                                 use_list=False,
                                 strict_map_key=False,
                                 object_hook=msgpack_convert)
            setattr(temp_stats, attribute, stat)
        self.stat_calculator.merge_into(temp_stats)

    def dump(self):

        """Method to dump all results into the database"""

        refmap = msgpack.dumps(self.gene_matches, default=msgpack_default, strict_types=False)
        simplified = self.stat_calculator.serialize()
        stats = []
        for attribute in simplified.attributes:
            stat = msgpack.dumps(getattr(simplified, attribute),
                          default=msgpack_default, strict_types=True)
            stats.append((attribute, stat))

        return refmap, stats

    def add_to_refmap(self, result: ResultStorer) -> None:
        """
        :param result: the result of the compare function

        This method verifies whether the comparison in input is the best available for the
        candidate reference gene, and loads it into the refmap dictionary if such is the case.

        """

        if result is not None and result.ccode != ("u",):
            gid, tid = result.ref_gene[0], result.ref_id[0]
            if gid not in self.gene_matches:
                self.gene_matches[gid] = dict()
            if tid not in self.gene_matches[gid]:
                self.gene_matches[gid][tid] = list()
            self.gene_matches[gid][tid].append(result)
        return

    @classmethod
    def find_neighbours(cls, keys, position, distance=2000):
        """
        This class method is used to find the possible matches of a given prediction key.
        :param keys: the start
        :type keys: Mikado.utilities.intervaltree.IntervalTree

        :param position: the position of my prediction in the genome
        :type position: (int, int)

        :param distance: optional maximum distance of a prediction
        from reference before being called an unknown
        :type distance: int

        :return: a list of distances, with the following format:
                 (start, end), distance
        """

        start, end = position

        # This should happen only if we are analysing a prediction from a scaffold
        # with no reference annotation on it
        if len(keys) == 0:
            return []

        found = keys.search(start, end, max_distance=distance)

        distances = []
        for key in found:
            # Append the key (to be used later for retrieval) and the distance
            distances.append(
                ((key.start, key.end),
                 max(0,
                     max(start, key.start) - min(end, key.end))))

        # Sort by distance
        distances = sorted(distances, key=operator.itemgetter(1))

        return distances

    @staticmethod
    def dubious_getter(dubious_result):
        """
        Function used to perform the sorting of the matches.
        :param dubious_result: a result
        :type dubious_result: ResultStorer
        """
        getter = operator.attrgetter("j_f1", "n_f1")
        return getter(dubious_result[0])

    @staticmethod
    def get_f1(curr_result):
        """
        Simple getter for F1 statistics (N_F1, J_F1 and distance).
        :param curr_result: a result storer
        :type curr_result: ResultStorer

        :return: (J_F1, N_F1, distance)
        :rtype (float, float, int)
        """
        return (curr_result.j_f1[0],
                curr_result.n_f1[0],
                -curr_result.distance[0],
                (curr_result.ccode[0] not in ("x", "X")))

    def __prepare_transcript(self, prediction: Transcript):
        """
        Private method that checks that a prediction transcript is OK
        before starting to analyse its concordance with the reference.
        :param prediction:
        :return:
        """

        # Prepare the prediction to be analysed
        prediction.logger = self.logger
        try:
            prediction.finalize()
            # noinspection PyUnresolvedReferences
            if self.args.exclude_utr is True:
                prediction.remove_utrs()
        except InvalidCDS:
            try:
                prediction.strip_cds()
            except InvalidTranscript as err:
                self.logger.warning("Invalid transcript (due to CDS): %s", prediction.id)
                self.logger.warning("Error message: %s", err)
                # self.done += 1
                self.print_tmap(None)
                return None
        except InvalidTranscript as err:
            self.logger.warning("Invalid transcript: %s", prediction.id)
            self.logger.warning("Error message: %s", err)
            # self.done += 1
            self.print_tmap(None)
            return None
        return prediction

    def __check_for_fusions(self, prediction, matches, fuzzymatch=None):

        """
        This private method checks whether a transcript with
        multiple matches at distance 0 is indeed a fusion, or not.
        :param prediction:
        :param matches:
        :type matches: list
        :return:
        """

        # Input: dict[(start, end), distance]: [gene1, gene2 ...]
        # What we need to achieve in this section of the program
        # Detect which genes in range *are* fused
        # Assign a "fused" tag to each transcript in the fused gene that passes the criteria

        results = []
        new_matches = dict()
        match_to_gene = dict()

        strands = collections.defaultdict(set)

        # Get all the results for the single gene
        # We *want* to do the calculation for all hits
        for match in matches:
            match_to_gene[match[0]] = self.positions[prediction.chrom][match[0]]
            for gene, gene_match in iter((gene, self.genes[gene]) for gene in
                                         match_to_gene[match[0]]):
                strands[gene_match.strand].add(gene)
                new_matches[gene] = sorted(
                    [self.calc_and_store_compare(prediction, tra, fuzzymatch=fuzzymatch) for tra in gene_match],
                    key=self.get_f1, reverse=True)

        # If we have candidates for the fusion which are on its same strand
        # Keep only those. Otherwise in compact genomes we might call as "fusions"
        # all transcripts which are assigned to genes overlapping on the same strand!
        if prediction.strand is not None and len(strands[prediction.strand]) > 0:
            for strand in iter(key for key in strands.keys() if key != prediction.strand):
                for gene in strands[strand]:
                    # Add to the final results the data for the gene
                    results.extend(new_matches[gene])
                    del new_matches[gene]

        assert len(new_matches) > 0, "Filtered all results for {}. This is wrong!".format(prediction.id)

        fused = collections.defaultdict(list)
        dubious = set()
        best = set()

        for match, genes in match_to_gene.items():
            local_best = []
            for gene in genes:
                if gene not in new_matches:
                    continue
                elif new_matches[gene] == []:
                    continue
                # The best match is the first
                if new_matches[gene][0].j_f1[0] == 0 and new_matches[gene][0].n_recall[0] < 10:
                    dubious.add(gene)
                    continue
                else:
                    local_best.append(new_matches[gene][0])
                    fused[match].append(gene)
            if len(local_best) > 0:
                best.add((match, sorted(local_best, key=self.get_f1, reverse=True)[0]))

        if len(best) == 0:
            self.logger.debug("Filtered out all results for %s, selecting the best dubious one",
                              prediction.id)
            # I have filtered out all the results,
            # because I only match partially the reference genes
            best_result = sorted([new_matches[gene][0] for gene in dubious],
                                 key=self.get_f1, reverse=True)[0]

        else:
            best = [_[1] for _ in sorted(best, key=lambda res: (res[0][0], res[0][1]))]
            if len(best) == 1:
                best_result = best.pop()
            elif len(best) > 1 and self.__report_fusions is False:
                best_result = sorted(best, key=operator.attrgetter("j_f1", "n_f1")).pop()
            else:  # We have a fusion
                # Now retrieve the results according to their order on the genome
                # Keep only the result, not their position

                chrom = prediction.chrom
                start = min([prediction.start] + [self.genes[_.ref_gene[0]][_.ref_id[0]].start for _ in best])
                end = max([prediction.end] + [self.genes[_.ref_gene[0]][_.ref_id[0]].end for _ in best])
                location = "{}:{}..{}".format(chrom, start, end)

                best_result = []
                for result in best:
                    new_result = []
                    for key in ResultStorer.__slots__:
                        if key == "location":
                            new_result.append(location)
                        elif key == "ccode":
                            tup = tuple(["f"] + [getattr(result, key)[0]])
                            new_result.append(tup)
                        else:
                            new_result.append(getattr(result, key))
                    new_result = ResultStorer(*new_result)
                    best_result.append(new_result)

                for match, genes in fused.items():
                    for gene in iter(_ for _ in genes if _ not in dubious):
                        for position, result in enumerate(new_matches[gene]):
                            if result.j_f1[0] > 0 or result.n_recall[0] > 10:
                                result.ccode = ("f", result.ccode[0])
                                new_matches[gene][position] = result

        # Finally add the results
        for gene in new_matches:
            results.extend(new_matches[gene])
        self.logger.debug("\n".join([str(result) for result in results]))

        return results, best_result

    def __prepare_result(self, prediction, distances, fuzzymatch=0):

        """
        This private method prepares the matching result for cases where the
        minimum distance from a reference gene/transcript is 0.
        :param prediction:
        :param distances:
        :return:
        """

        matches = list(distance for distance in distances if distance[1] == 0)
        self.logger.debug("Matches for %s: %s", prediction.id,
                          ", ".join([_.id for _ in self.genes[self.positions[prediction.chrom][distances[0][0]][0]]]))

        same_strand = False
        if len(matches) > 1 and prediction.strand is not None:
            correct = list()
            for match in matches:
                for gid in self.positions[prediction.chrom][match[0]]:
                    if any([self.genes[gid].strand in (None, prediction.strand)]):
                        correct.append(match)
                        break
            if len(correct) > 0:
                matches = correct[:]
                same_strand = True
            del correct
        elif len(matches) > 1 and prediction.strand is None:
            same_strand = True

        if len(matches) > 1 and same_strand is True:
            self.logger.debug("More than one match for %s: %s",
                              prediction.id,
                              matches)
            results, best_result = self.__check_for_fusions(prediction, matches, fuzzymatch=fuzzymatch)
        else:
            matches = [self.genes[_] for _
                       in self.positions[prediction.chrom][matches[0][0]]]
            results = []
            for match in matches:
                self.logger.debug("%s: type %s", repr(match), type(match))
                for tra in match:
                    try:
                        results.append(self.calc_and_store_compare(prediction, tra, fuzzymatch=fuzzymatch))
                    except TypeError:
                        failed = []
                        for intron in prediction.introns:
                            if not (isinstance(intron[0], int) and isinstance(intron[1], int)):
                                failed.append(intron)

                        raise TypeError((failed, type(prediction), prediction.introns))

            results = sorted(results, reverse=True,
                             key=self.get_f1)

            best_result = results[0]

        return results, best_result

    def self_analyse_prediction(self, prediction: Transcript, distances, fuzzymatch=0):

        """This method will be invoked during a self analysis run."""

        assert len(distances) >= 1 and distances[0][1] == 0

        genes = []
        __gene_removed = False
        __gene_found = False
        for distance in distances:
            for posis in self.positions[prediction.chrom][distance[0]]:
                gene = self.genes[posis]
                if prediction.id in gene.transcripts.keys():
                    __gene_found = True
                    if len(gene.transcripts) == 1:
                        __gene_removed = True
                        continue
                genes.append((self.genes[posis], distance[1]))

        if __gene_found is False:
            raise AssertionError("Parent for {} not found in the index!".format(prediction.id))

        genes = sorted(genes, key=operator.itemgetter(1))

        # noinspection PyUnresolvedReferences
        if len(genes) == 0 or genes[0][1] > self.args.distance:
            # noinspection PyTypeChecker,PyUnresolvedReferences
            best_result = ResultStorer("-", "-",
                                       "u",
                                       prediction.id,
                                       prediction.parent[0],
                                       prediction.exon_num,
                                       "-",
                                       *[0] * 9 + ["-"] + [prediction.location])
            self.stat_calculator.store(prediction, best_result, None)
            results = [best_result]
        elif genes[0][1] > 0:  # Up or downstream
            results = [self.calc_and_store_compare(prediction, reference, fuzzymatch=fuzzymatch)
                       for reference in genes[0][0]]
            best_result = sorted(results,
                                 key=operator.attrgetter("distance"))[0]
        else:
            # All other cases
            genes = [_[0] for _ in genes if _[1] == 0]
            same_strand = False
            if prediction.strand is not None:
                same_strand_genes = [_ for _ in genes if _.strand == prediction.strand]
                if len(same_strand_genes) >= 1:
                    genes = same_strand_genes
                    same_strand = True

            if len(genes) == 1:
                results = [self.calc_and_store_compare(prediction, reference, fuzzymatch=fuzzymatch)
                           for reference in genes[0] if reference.id != prediction.id]
                assert len(results) > 0, (genes[0].transcripts.keys(), prediction.id)
                best_result = sorted(results,
                                     key=operator.attrgetter("distance"))[0]
            else:
                result_dict = dict()
                results = []

                for gene in genes:
                    if gene.id in prediction.parent:
                        if prediction.id not in gene.transcripts:
                            raise AssertionError("{} not found in {}!".format(prediction.id, gene.id))
                        elif len(gene.transcripts) == 1:
                            raise AssertionError(
                                "We should have ignored {}! Gene removed: {}".format(
                                    gene.id, __gene_removed))

                    result_dict[gene.id] = sorted(
                        [self.calc_and_store_compare(prediction, reference, fuzzymatch=fuzzymatch)
                         for reference in gene if reference.id != prediction.id],
                        key=self.get_f1,
                        reverse=True)
                    if len(result_dict[gene.id]) == 0:
                        raise AssertionError(
                            "Nothing found for {} vs. {} (transcripts: {})".format(
                                gene.id, prediction.id, ", ".join(list(gene.transcripts.keys()))
                            ))

                if same_strand is True:
                    # This is a fusion, period
                    results = []
                    best_results = []
                    for gene in result_dict:
                        if result_dict[gene][0].n_recall[0] > 10 or result_dict[gene][0].j_f1[0] > 0:
                            results.extend(result_dict[gene])
                            best_results.append(result_dict[gene][0])
                    if len(best_results) == 1:
                        best_result = best_results[0]
                    elif len(best_results) > 1 and self.__report_fusions is True:
                        values = []
                        for key in ResultStorer.__slots__:
                            if key in ["gid", "tid", "distance", "tid_num_exons"]:
                                values.append(getattr(best_results[0], key))
                            elif key == "ccode":
                                values.append(tuple(["f"] + [_.ccode[0] for _ in best_results]))
                            else:
                                val = tuple([getattr(result, key)[0] for result in best_results])
                                values.append(val)
                        best_result = ResultStorer(*values)
                        for result in results:
                            if result.j_f1[0] > 0 or result.n_recall[0] > 10:
                                result.ccode = ("f", result.ccode[0])
                    elif len(best_results) > 1 and self.__report_fusions is False:
                        best_result = [_ for _ in sorted(best_results, key=operator.attrgetter("j_f1", "n_f1",))].pop()

                # Check how many
                if not results:
                    # This is not a fusion. Let us just select the best match.
                    best = None
                    for gene in result_dict:
                        if best is None:
                            best = [gene, result_dict[gene][0]]
                        else:
                            if sorted([best[1], result_dict[gene][0]],
                                      key=self.get_f1, reverse=True)[0] == result_dict[gene][0]:
                                best = [gene, result_dict[gene][0]]
                    results = result_dict[best[0]]
                    best_result = best[1]

        return results, best_result

    def get_best(self, prediction: Transcript, fuzzymatch=None):

        """
        :param prediction: the candidate transcript to be analysed
        :type prediction: Transcript

        :param fuzzymatch: leniency in determining whether two introns are related
        :type fuzzymatch: (int|None)

        This function will get the best possible assignment for each transcript.
        Fusion genes are called when the following conditions are verified:
        - the prediction intersects (at least) two transcripts in (at least)
        two different loci
        - the suspected fusion transcript lies on the same strand of all
        candidate fused genes
        - each candidate transcript has at least one fusion or
        10% of its nucleotides covered by the fusion transcript.
        The 10% threshold is hard-coded in the function.
        """

        self.logger.debug("Started with %s", prediction.id)

        if fuzzymatch is None:
            fuzzymatch = self.__fuzzymatch

        # Quickly check that the transcript is OK
        prediction = self.__prepare_transcript(prediction)
        if prediction is None:
            return None
        elif prediction.finalized is False:
            self.logger.error("%s failed to be finalised. Ignoring it.", prediction.id)
            return None

        # Ignore non-coding RNAs if we are interested in protein-coding transcripts only
        # noinspection PyUnresolvedReferences
        if self.args.protein_coding is True and prediction.combined_cds_length == 0:
            self.logger.debug("No CDS for %s. Ignoring.", prediction.id)
            # self.done += 1
            self.print_tmap(None)
            return None

        if prediction.chrom in self.indexer:
            keys = self.indexer[prediction.chrom]
        else:
            keys = IntervalTree()

        # noinspection PyUnresolvedReferences
        # Here I am getting all the hits within striking distance for my hit
        distances = self.find_neighbours(keys,
                                         (prediction.start, prediction.end),
                                         distance=self.args.distance)
        if self.self_analysis is True:
            results, best_result = self.self_analyse_prediction(prediction, distances,
                                                                fuzzymatch=fuzzymatch)
        else:
            self.logger.debug("Distances for %s: %s", prediction.id, distances)
            # Unknown transcript
            # noinspection PyUnresolvedReferences
            if len(distances) == 0 or distances[0][1] > self.args.distance:
                ccode = "u"
                # noinspection PyTypeChecker,PyUnresolvedReferences
                best_result = ResultStorer("-", "-",
                                           ccode,
                                           prediction.id,
                                           prediction.parent[0],
                                           prediction.exon_num,
                                           "-",
                                           *[0] * 9 + ["-"] + [prediction.location])
                self.stat_calculator.store(prediction, best_result, None)
                results = [best_result]
            elif distances[0][1] > 0:
                # Polymerase run-on
                match = self.genes[self.positions[prediction.chrom][distances[0][0]][0]]
                results = [self.calc_and_store_compare(prediction, reference) for reference in match]
                best_result = sorted(results, key=operator.attrgetter("distance"))[0]
            else:
                # All other cases
                try:
                    results, best_result = self.__prepare_result(prediction, distances, fuzzymatch=fuzzymatch)
                except (InvalidTranscript, ZeroDivisionError, AssertionError) as exc:
                    self.logger.error("Something went wrong with %s. Ignoring it. Error: %s",
                                      prediction.id, exc)
                    self.logger.debug(exc)
                    return None

        if self.use_prediction_alias is True:
            if isinstance(best_result, list):
                for idx in range(len(best_result)):
                    best_result[idx].tid = prediction.alias
            else:
                best_result.tid = prediction.alias
            for idx in range(len(results)):
                results[idx].tid = prediction.alias

        for result in results:
            self.add_to_refmap(result)

        self.logger.debug("Finished with %s", prediction.id)
        if isinstance(best_result, list):
            for result in best_result:
                self.print_tmap(result)
        else:
            self.print_tmap(best_result)

        return best_result

    def finish(self):
        """
        This function is called upon completion of the input file parsing.
        It will print out the reference file assignments into the refmap
        file, and the final statistics into the stats file.
        """

        self.logger.debug(f"Finished parsing, total: {self.done} transcript{'s' if self.done > 1 else ''}.")
        self.print_refmap()
        self.stat_calculator.print_stats()
        self.logger.info("Finished printing final stats")
        self.tmap_out.close()
        self.logger.info("Closed output files")

    def calc_and_store_compare(self, prediction: Transcript, reference: Transcript, fuzzymatch=0) -> ResultStorer:
        """Thin layer around the calc_and_store_compare class method.

        :param prediction: a Transcript instance.

        :param reference: a Transcript instance to which the prediction is compared to.

        :rtype ResultStorer

        """

        result, reference_exon = c_compare(prediction, reference,
                                           lenient=self.lenient, fuzzymatch=fuzzymatch)

        assert reference_exon is None or reference_exon in reference.exons
        self.stat_calculator.store(prediction, result, reference_exon)

        return result

    @staticmethod
    def compare(prediction: Transcript,
                reference: Transcript,
                lenient=False,
                strict_strandedness=False,
                fuzzy_match=0) -> (ResultStorer, tuple):

        """Function to compare two transcripts and determine a ccode. Thin wrapper
        around the compare cython code.

        :param prediction: the transcript query
        :type prediction: Transcript

        :param reference: the reference transcript against which we desire to
        calculate the ccode and other stats.
        :type reference: Transcript

        :param lenient: boolean flag. If switched to True, the comparison will be performed by considering
        only the *internal* boundaries of the transcripts.
        :type lenient: bool

        :param strict_strandedness: boolean flag. If switched to True, transcripts without a strand
        will be considered as stranded when calculating the type of overlap with the transcript.
        Eg. if the option is set to True and the transcript has a class code of e, it will become an x.
        :type strict_strandedness: bool

        :rtype (ResultStorer, (int,int)) | (ResultStorer, None)

        Please see the class_codes subpackage for details on the available class codes.

        This is a class method, and can therefore be used outside of a class instance.
        """

        return c_compare(prediction,
                         reference,
                         lenient=lenient,
                         strict_strandedness=strict_strandedness,
                         fuzzymatch=fuzzy_match)

    def print_tmap(self, res):
        """
        This method will print a ResultStorer instance onto the TMAP file.
        :param res: result from compare
        :type res: (ResultStorer | None)
        """

        if res is not None:
            if not isinstance(res, ResultStorer):
                self.logger.exception("Wrong type for res: %s", str(type(res)))
                self.logger.exception(repr(res))
                raise ValueError
            else:
                if self.printout_tmap is False:
                    pass
                else:
                    self.__done += 1
                    self.tmap_rower.writerow(res.as_dict())

    @staticmethod
    def result_sorter(result):

        """
        Method to sort the results for the refmap. Order:
        - CCode does not contain "x", "P", "p" (i.e. fragments on opposite strand or
        polymerase run-on fragments)
        - Exonic F1 (e_f1)
        - Junction F1 (j_f1)
        - "f" in ccode (i.e. transcript is a fusion)
        - Nucleotide F1 (n_f1)

        :param result: a resultStorer object
        :type result: ResultStorer
        :return: (int, float, float, float)
        """

        bad_ccodes = ["x", "X", "P", "p"]
        bad_ccodes = set(bad_ccodes)

        orderer = (len(set.intersection(bad_ccodes, set(result.ccode))) == 0,
                   result.j_f1, result.e_f1,
                   result.n_f1,
                   "f" in result.ccode)

        return orderer

    def print_refmap(self) -> None:

        """Function to print out the best match for each gene."""
        self.logger.info("Starting printing RefMap")
        # noinspection PyUnresolvedReferences
        if self.args.gzip is False:
            opening_function = partial(open, "{0}.refmap".format(self.args.out), 'wt')
        else:
            opening_function = partial(gzip.open, "{0}.refmap.gz".format(self.args.out), 'wt')

        with opening_function() as out:
            if self.args.extended_refmap is True:
                fields = ["ref_id", "ccode", "tid", "gid",
                          "nRecall", "nPrecision", "nF1",
                          "jRecall", "jPrecision", "jF1",
                          "eRecall", "ePrecision", "eF1",
                          "ref_gene",
                          "best_ccode", "best_tid", "best_gid",
                          "best_nRecall", "best_nPrecision", "best_nF1",
                          "best_jRecall", "best_jPrecision", "best_jF1",
                          "best_eRecall", "best_ePrecision", "best_eF1",
                          "location"]
            else:
                fields = ["ref_id", "ccode", "tid", "gid",
                          "nF1", "jF1", "eF1",
                          "ref_gene",
                          "best_ccode", "best_tid", "best_gid",
                          "best_nF1", "best_jF1", "best_eF1",
                          "location"]
            out_tuple = namedtuple("refmap", fields)

            rower = csv.DictWriter(out, fields, delimiter="\t")
            rower.writeheader()

            for gid in sorted(self.gene_matches.keys()):

                rows = []
                best_picks = []
                assert len(self.gene_matches[gid].keys()) > 0
                for tid in sorted(self.gene_matches[gid].keys()):
                    if len(self.gene_matches[gid][tid]) == 0:
                        # First part of the tuple
                        row = tuple([tid, gid] + ["NA"] * 7)
                    else:
                        # Choose the best hit for the transcript
                        if any((x.j_f1[0] > 0 or x.n_f1[0] > 0) for x in self.gene_matches[gid][tid]):
                                best = sorted(self.gene_matches[gid][tid],
                                              key=self.result_sorter, reverse=True)[0]
                        else:
                            best = sorted(self.gene_matches[gid][tid],
                                          key=operator.attrgetter("distance"),
                                          reverse=False)[0]
                        best_picks.append(best)
                        # Store the result for the transcript
                        if self.args.extended_refmap is True:
                            row = tuple([tid, gid, ",".join(best.ccode),
                                         best.tid, best.gid,
                                         best.n_recall[0], best.n_prec[0], best.n_f1[0],
                                         best.j_recall[0], best.j_prec[0], best.j_f1[0],
                                         best.e_recall[0], best.e_prec[0], best.e_f1[0],
                                         best.location])
                        else:
                            row = tuple([tid, gid, ",".join(best.ccode),
                                         best.tid, best.gid,
                                         best.n_f1[0], best.j_f1[0], best.e_f1[0],
                                         best.location])

                    rows.append(row)

                if len(best_picks) > 0:
                    best_pick = sorted(best_picks,
                                       key=self.result_sorter,
                                       reverse=True)[0]
                else:
                    best_pick = None

                for row in rows:
                    if best_pick is not None:
                        if self.self_analysis is False:
                            assert row[2] != "NA", row
                        if self.args.extended_refmap is True:
                            row = out_tuple(row[0],  # Ref TID
                                            row[2],  # class code
                                            row[3],  # Pred TID
                                            row[4],  # Pred GID
                                            row[5], row[6], row[7],  # N
                                            row[8], row[9], row[10], #J
                                            row[11], row[12], row[13], #E
                                            row[1],
                                            ",".join(best_pick.ccode),
                                            best_pick.tid,
                                            best_pick.gid,
                                            best_pick.n_recall[0], best_pick.n_prec[0], best_pick.n_f1[0],
                                            best_pick.j_recall[0], best_pick.j_prec[0], best_pick.j_f1[0],
                                            best_pick.e_recall[0], best_pick.e_prec[0], best_pick.e_f1[0],
                                            row[14]  #Location
                                            )
                        else:
                            # fields = ["ref_id", "ccode", "tid", "gid",
                            #                           "nF1", "jF1", "eF1",
                            #                           "ref_gene",
                            #                           "best_ccode", "best_tid", "best_gid",
                            #                           "best_nF1", "best_jF1", "best_eF1",
                            #                           "location"]
                            try:
                                row = out_tuple(ref_id=row[0],  # Ref TID
                                                ccode=row[2],  # class code
                                                tid=row[3],  # Pred TID
                                                gid=row[4],  # Pred GID
                                                nF1=row[5], jF1=row[6], eF1=row[7],  # Pred F1
                                                ref_gene=row[1],
                                                best_ccode=",".join(best_pick.ccode),
                                                best_tid=best_pick.tid,
                                                best_gid=best_pick.gid,
                                                best_nF1=best_pick.n_f1[0],
                                                best_jF1=best_pick.j_f1[0],
                                                best_eF1=best_pick.e_f1[0],
                                                location=row[8]  # Location
                                                )
                            except IndexError:
                                self.logger.critical("Error in creating the refmap output")
                                self.logger.critical(row)
                                raise
                    else:
                        if self.args.extended_refmap is True:
                            row = out_tuple(*[row[0]] + ["NA"] * 12 + [row[1]] + ["NA"] * 12 + [
                                self.genes[gid][row[0]].location])
                        else:
                            row = out_tuple(*[row[0]] + ["NA"] * 6 + [row[1]] + ["NA"] * 6 + [
                                self.genes[gid][row[0]].location])
                    # noinspection PyProtectedMember,PyProtectedMember
                    rower.writerow(row._asdict())
        self.logger.info("Finished printing RefMap")
        return None
